#!/usr/bin/perl
#!c:/perl/bin/perl.exe

# GWebCache - Gnutella Web-Based Host Cache System v2
# Copyright (c) 2002-2003 by Hauke Dmpfling <haukex@zero-g.net>
#
# Please see this website for the license for this software:
# http://www.gnu.org/licenses/gpl.html
# For more information visit:
# http://gnucleus.sourceforge.net/gwebcache/
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#use KCatch; # catches critical errors (debugging)
use Storable; #qw(lock_store lock_retrieve); #DB
#use CGI::Carp qw(warningsToBrowser fatalsToBrowser); #DB

# ### User Vars ###
@allowed_nets = qw(gnutella gnutella2); # add your own networks here! (you must include "gnutella")
$data_file = 'gdata.dat';
$stat_log = 'gstatlog.txt';
$max_logsize = 2**19; # 0.5 MB

# It's better not to mess with the following settings!
$default_net = 'gnutella';
$max_hosts = 20;
$max_urls = 10;
$max_nets = 50;
$old_ping = 1; # set this to 0 when GWC2 is more widespread
# Security Settings
$security_time = 60*55; # seconds
@default_allow = qw(LIME GNUC BEAR GTKG RAZA ACQX MNAP SWAP MUTE TEST MRPH);
$use_advanced_security = 0; # take part in advanced security?
$allow_file = 'gsecurity.txt'; # used with advanced security only
$security_server = 'http://gnucleus.sourceforge.net/cgi-bin/gwcsec.cgi'; # leave at default server
# ###


# DEBUG OPTIONS ONLY - DON'T USE
$debug_no_security=0; # Completely disable security checks
$debug_pretend_security=0; # Do security checks but don't fail
$debug_out_html=0; # Output responses as text/html inside <PRE> tags (needed for some browsers)

# Constants
*VERSION = \'GWebCache 1.0.0';
$now = time;

# Init
$file_read=0;
$output_out=0;
@output=();
@loginfo=();
$rip=$ENV{REMOTE_ADDR};


# Nuke any server that doesn't give us REMOTE_ADDR
unless( $rip ) {
	print "Content-type: text/plain\n\n";
	print "ERROR: Server does not support REMOTE_ADDR, this cache shouldn't be running here\n";
	exit;
}

# ### Process Requests ###
&parse_form;

# Hack: gotta get rid of very bad clients until they learn and update!
# Note: MMMM will not re-request as often if we give it a reply that looks ok.
if( $form{client} eq 'MMMM' && $form{version} =~ /^[12]/ ) {
	print "Content-type: text/plain\n\n";
	if( $form{urlfile} ) {
		print "http://www.morpheus-os.com\n";
	}
	elsif( $form{hostfile} ) {
		print "140.99.15.144:80\n";
	}
	exit;
}

# Read data file
&read_file;

# Security Checks
if( scalar %form ) { # Any kind of requests
	$form{client}.=$form{version} if $form{version}; # Old-skool client/version info stuff
	
	unless( $debug_no_security ) { #Debug
	
	unless( $form{client}=~/^[A-Z]{4}.{1,16}$/ ) {
		push(@loginfo,'Err:ClientInfo');
		&error('Must send correctly formatted client version information');
	}
	
	&check_security_updates if $use_advanced_security;
	
	my $cl = substr($form{client},0,4);
	unless( grep($cl eq $_,@default_allow) ) {
		push(@loginfo,'Err:AccessDenied');
		&error('Access denied by ACL. Please e-mail gwc@zero-g.net if you ');
	}

	my $found=0;
	if( @{$data{_sec}} ) {
		for my $i (@{$data{_sec}}-1..0) { # reverse order so delete works
			if( $now > $data{_sec}[$i]{t}+$security_time ) {
				splice @{$data{_sec}}, $i, 1;
			}
			elsif( $data{_sec}[$i]{h} eq $rip ) {
				$found=1;
				if( $data{_sec}[$i]{c}>1 ) {
					push(@loginfo,'Err:Early');
					&error('You\'re only allowed one request per hour!');
				}
				else {
					$data{_sec}[$i]{c}++;
				}
			}
		}
	}
	unshift(@{$data{_sec}},
		{	t => $now,
			h => $rip,
			c => 1	}) unless $found;
	
	} #/Debug
}
else { # No requests, default page
	&htmlpage;
	exit;
}

# Multi-Network Support
$getaltnets=0;
if( $form{net} ) {
	$net = lc($form{net});
	# Are we supporting this net?
	if( grep(lc eq $net,@allowed_nets) ) {
		push(@loginfo,"Net:$net");
	}
	else {
		$getaltnets=1;
		push(@loginfo,"AltNet:$net");
		push(@output,'I|net-not-supported');
	}
}
else {
	$net = $default_net;
}

# Get Request
if( $form{get} ) {
	push(@loginfo,'Get');
	if( $getaltnets ) { # Get Data for alternate network
		my @nets = keys %{$data{_nets}};
		foreach my $tnet (@nets) {
			if( $tnet eq $net ) {
				my $age = $now - $data{_nets}{$tnet}{t};
				push(@output,"U|$data{_nets}{$tnet}{u}|$age");
			}
		}
	}
	else { # Normal get
		foreach my $host (@{$data{$net}{hosts}}) {
			my $age = $now - $host->{t};
			push(@output,"H|$host->{h}|$age|$host->{c}");
		}
		foreach my $url (@{$data{$net}{urls}}) {
			my $age = $now - $url->{t};
			push(@output,"U|$url->{u}|$age");
		}
	}
}
else { # Any other shiz?
	my $out='';
	if( $form{hostfile} ) { # Old-skool hostfile
		push(@loginfo,'Hostfile');
		if( $getaltnets ) { # Unsupported network?
			$out="\n";
		}
		else {
			foreach my $host (@{$data{$net}{hosts}}) {
				#if( $form{cluster} ) { # Support clustering even on old requests
				#	$out.="$host->{h}|$host->{c}\n";
				#}
				#else {
					$out.="$host->{h}\n";
				#}
			}
		}
		&output($out);
		&finish_up;		
	}
	if( $form{urlfile} ) { # Old-skool urlfile
		push(@loginfo,'Urlfile');
		if( $getaltnets ) { # Unsupported network?
			my @nets = keys %{$data{_nets}};
			foreach my $tnet (@nets) {
				if( $tnet eq $net ) {
					$out.="$data{_nets}{$tnet}{u}\n";
				}
			}
		}
		else {
			foreach my $url (@{$data{$net}{urls}}) {
					$out.="$url->{u}\n";
			}
		}
		&output($out);
		&finish_up;
	}
}

# Update request
if( $form{ip} || $form{url} ) {
	$upd_out='';
	&update;
	unless( $form{update} || $form{get} ) { # Give old-skool response to old-skool updates
		&output("OK\n$upd_out");
		&finish_up;
	}
}

# Ping request
if( $form{ping} ) {
	push(@loginfo,'Ping');
	if( $old_ping && !($form{get} || $form{update} || $form{net} || $form{cluster}) ) {
		&output("PONG $VERSION");
		&finish_up;
	}
	else {
		unshift(@output,"I|pong|$VERSION");
	}
}

unshift(@output,'I|nothing')
	unless @output;

# Do Output
#map {tr/|/\t/} @output; # Tabs instead of |
&output(join("\n",@output)."\n");
&finish_up;


# ## Subs

# ----- UPDATE -----
sub update {
	my $sip=$form{ip};
	my $surl=$form{url};
	
	push(@loginfo,"Upd:$sip/$surl");
	
	# Check Submissions
	unless( $debug_no_security ) { #Debug
	if( $sip ) {
		unless( $sip=~/^$rip\:(\d{1,5})$/ && $1>0 && $1<2**16 ) {
			$sip=undef
				unless $debug_pretend_security; #Debug
			unshift(@output,'I|update|WARNING|Rejected IP');
			$upd_out.="WARNING: Rejected IP\n"; # Old-skool output
			push(@loginfo,'ErrUpd:RejectIP');
		}
	}
	if( $surl ) {
		unless( $surl=~/^http\:\/\/.+$/i ) {
			$surl=undef
				unless $debug_pretend_security; #Debug
			unshift(@output,'I|update|WARNING|Rejected URL');
			$upd_out.="WARNING: Rejected URL\n"; # Old-skool output
			push(@loginfo,'ErrUpd:RejectURL');
		}
	}
	} #/Debug
	
	# Taking submission for alternate cache list
	if( $getaltnets && $surl ) {
			my $found=0;
			my @nets = keys %{$data{_nets}};
			foreach my $tnet (@nets) {
				if( $tnet eq $net ) {
					$data{_nets}{$tnet}{u} = $surl;
					$data{_nets}{$tnet}{t} = $now;
					$found=1;
				}
			}
			$data{_nets}{$net} = {	'u'=>$surl,	't'=>$now	} unless $found; # New
			@nets = keys %{$data{_nets}};
			if( @nets>$max_nets ) { # Purge older net data
				@nets = sort { 
						$data{_nets}{$a}{t} <=> $data{_nets}{$b}{t} # Sort by age
					} @nets;
				for my $i (0..@nets-$max_nets-1) {
					delete $data{_nets}{$nets[$i]}; # Delete oldest
				}
			}
	}
	
	# Normal Submission
	else {
		# Host Update
		if( $sip ) {
			
			# First take care of cluster stuff, if it's included in the request
			my $cf=undef;
			if( $form{cluster} ) {
				$cf = $form{cluster};
				my @ckeys = split(/,/,$cf);
				$cf='';
				foreach $key (@ckeys) {
					$key=~s/[^A-Za-z0-9\.\-\_\:]//g;
					$cf.=',' if $cf;
					$cf.=$key;
					last if length($cf)>256;
				}
			}
			
			my $found=0;
			my ($sip2) = $sip=~/^(.+)\:.+$/;
			foreach my $host (@{$data{$net}{hosts}}) {
				if( $host->{h}=~/^$sip2/ ) {
					$host->{h}=$sip;
					$host->{t}=$now;
					$host->{c}=$cf;
					$found=1;
					last;
				}
			}
			unshift(@{$data{$net}{hosts}},
				{	'h'=>$sip,
					't'=>$now,
					'c'=>$cf	}) unless $found; # New
			@{$data{$net}{hosts}} = @{$data{$net}{hosts}}[0..$max_hosts-1]
					if @{$data{$net}{hosts}}>$max_hosts; # Overflow
		}
		
		# URL Update
		if( $surl ) {
			my $found=0;
			foreach my $url (@{$data{$net}{urls}}) {
				if( $url->{u} eq $surl  ) {
					$url->{u}=$surl;
					$url->{t}=$now;
					$found=1;
					last;
				}
			}
			unshift(@{$data{$net}{urls}},
				{	'u'=>$surl,
					't'=>$now	}) unless $found; # New
			@{$data{$net}{urls}} = @{$data{$net}{urls}}[0..$max_urls-1]
					if @{$data{$net}{urls}}>$max_urls; # Overflow
		}
	}
	
	unshift(@output, 'I|update|OK');
}

# ----- READ_FILE -----
sub read_file {
	return if $file_read;
	$file_read=1;
	
	my $lock_mode = 2; # 1=Shared, 2=Exclusive, 8=Unlock
	
	# Lock File
	open(LOCK, $data_file.'.lock') || &cgi_die("Locking file error: $!");
	flock(LOCK, $lock_mode) || &cgi_die("File locking error: $!");
	
	# Load File
	eval {
		#%data=%{ lock_retrieve($data_file) }; #DB
		%data=%{ retrieve($data_file) };
	};
	
	# In case file was corrupted
	unless( %data ) {
		push(@loginfo,'Warn:DataReset');
		$data{$default_net}{hosts} = [];
		$data{$default_net}{urls} = [];
		$data{_sec} = [];
		$data{_nets} = {};
	}
	
}

# ----- FINISH_UP -----
sub finish_up {
	&dolog;
	# Save File
	#lock_store(\%data, $data_file) || &cgi_die("Data file save error: $!"); #DB
	store(\%data, $data_file) || &cgi_die("Data file save error: $!");
	close(LOCK);
	exit;
}

# ----- DOLOG -----
sub dolog {
	my $cl = $form{client};
	
	if( (-s $stat_log) > $max_logsize ) { # Clear logfile when too big
		open(LOG,'>'.$stat_log);
	}
	else {
		open(LOG,'>>'.$stat_log);
	}
	flock(LOG,2);
	print LOG '['.gmtime(time)."] $rip/$cl REQ ".join(', ',@loginfo)."\n";
	close(LOG);
}

# ----- CGI_DIE -----
sub cgi_die {
	my $msg = shift;
	&output("ERROR: $msg");
	&finish_up;
}

# ----- ERROR -----
sub error {
	my $msg = shift;
	if( $debug_pretend_security ) { #Debug
		push(@output,"D|ERROR|$msg");
	}
	else {
		&cgi_die($msg);
	}
}

# ----- security -----

sub check_security_updates {
	use LWP::UserAgent;
	
	if( (-M $allow_file) > 0.5 ) { # Allow File is old
		push(@loginfo,'UpdateSec');
		my $ua = LWP::UserAgent->new(timeout => 30);
		my $req = HTTP::Request->new('GET', $security_server."?version=$VERSION");
		my $res = $ua->request($req);
		if( $res->is_success && $res->content=~/^[\s\n\r]*([A-Z\ ]+)[\s\n\r]*$/ ) {
			my $secup = $1;
			open(FILE,'>'.$allow_file);
			flock(FILE,2);
			print FILE "$secup\n";
			close(FILE);
			@default_allow = split(/ /,$secup);
			#push(@output,'D|updated security'); #DB
		}
		else {
			open(FILE,'>>'.$allow_file);
			flock(FILE,2);
			print FILE ' ';
			close(FILE);
		}
	}
	else {
		open(FILE,$allow_file);
		flock(FILE,1);
		my $secup=<FILE>;
		close(FILE);
		if( $secup=~/^[\s\n\r]*([A-Z\ ]+)[\s\n\r]*$/ ) {
			@default_allow = split(/ /,$1);
		}
	}
	#push(@output,"D|@default_allow"); #DB
	
}

# ----- OUTPUT -----
sub output {
	my $txt = shift;
	unless( $output_out ) {
		print "Pragma: no-cache\n";
		print "Cache-Control: no-store, no-cache, must-revalidate\n";
		print "Cache-Control: post-check=0, pre-check=0\n";
		print "X-Remote-IP: $rip\n";
		if( $debug_out_html ) {
			print "Content-type: text/html\n\n";
		}
		else {
			print "Content-type: text/plain\n\n";
		}
		$output_out=1;
	}
	if( $debug_out_html ) {
		print "<PRE>$txt</PRE>";
	}
	else {
		print "$txt";
	}
	#warningsToBrowser(1); #DB
}

# ----- PARSE_FORM -----
sub parse_form {
	my (@pairs,$buffer,$pair);
	if ($ENV{REQUEST_METHOD} eq 'GET') {
		@pairs = split(/&/, $ENV{QUERY_STRING});
	}
	elsif ($ENV{REQUEST_METHOD} eq 'POST') {
		read(STDIN, $buffer, $ENV{CONTENT_LENGTH});
		@pairs = split(/&/, $buffer);
	}
	foreach $pair (@pairs) {
		my ($name, $value) = split(/=/, $pair, 2);
		$value =~ tr/+/ /;
		$value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack('C', hex($1))/eg;
		$value =~ s/\|//g; # Protect us from evildoers
		$name =~ s/[^A-Za-z0-9\.\-\_]//g;
		$form{$name} = $value;
	}
}

# ----- HTMLPAGE -----
sub htmlpage {
	
	# Get Data
	&read_file;
	
	# Output Everything
	print "Content-type: text/html\n\n";
	print <<"(END HEAD HTML)";
<HTML>
<HEAD>
	<TITLE>GWebCache</TITLE>
	<META NAME="robots" CONTENT="noindex,nofollow">
<STYLE TYPE="text/css">
BODY,P,TD  {
	font-family: Tahoma, Arial, Helvetica, sans-serif;
	font-size: 10pt;
}
H1 { font-size: 16pt; }
H3 { font-size: 12pt; }
TD { font-size: 8pt; white-space: nowrap; background: lightblue; }
TH { font-size: 8pt; font-weight: bold; white-space: nowrap; background: lightsteelblue; }
A { color: blue; text-decoration: none; }
A:Hover { text-decoration: underline; }
A:Visited { color:darkblue; }
</STYLE>
</HEAD>
<BODY>
<H1>GWebCache</H1>

<P><B>
This is a GWebCache web-based host cache script for peer-to-peer networks.
</B></P>

<P>
Running: $ENV{SERVER_SOFTWARE} on $ENV{SERVER_NAME}<BR>
$VERSION<BR>
by Hauke D&auml;mpfling<BR>
visit <A HREF="http://gnucleus.sourceforge.net/">gnucleus.sourceforge.net</A>
</P>

<P>
<A HREF="http://gnucleus.sourceforge.net/gwebcache/"><IMG SRC="http://gnucleus.sourceforge.net/gwebcache/version_1.0.0.gif" ALT="Version Information" WIDTH=200 HEIGHT=100 BORDER=0></A>
</P>

<H3>Data</H3>

(END HEAD HTML)
	
	foreach my $anet (@allowed_nets) {
		print qq{<P><TABLE BORDER=0 CELLPADDING=3 CELLSPACING=2>\n};
		print qq{<TR><TH COLSPAN=2>Network: $anet</TH></TR>\n};
		print qq{<TR><TH>Host / Cluster Info</TH><TH>Age</TH></TR>\n};
		foreach my $host (@{$data{$anet}{hosts}}) {
			my $age = getage($now-$host->{t});
			my $wcluster =  $host->{c}?" / $host->{c}":'';
			print qq{<TR><TD><A HREF="http://$host->{h}">$host->{h}</A>$wcluster</TD><TD>$age</TD></TR>\n};
		}
		print qq{<TR><TH>Cache URL</TH><TH>Age</TH></TR>\n};
		foreach my $url (@{$data{$anet}{urls}}) {
			my $age = getage($now-$url->{t});
			print qq{<TR><TD><A HREF="$url->{u}">$url->{u}</A></TD><TD>$age</TD></TR>\n};
		}
		print qq{</TABLE></P>\n};
	}
	
	my @nets = keys %{$data{_nets}};
	print qq{<P><TABLE BORDER=0 CELLPADDING=3 CELLSPACING=2>\n};
	print qq{<TR><TH COLSPAN=3>Alternate Networks</TH></TR>\n};
	print qq{<TR><TH>Network</TH><TH>Cache URL</TH><TH>Age</TH></TR>\n};
	foreach my $anet (@nets) {
		my $age = getage($now-$data{_nets}{$anet}{t});
		print qq{<TR><TD>$anet</TD><TD><A HREF="$data{_nets}{$anet}{u}">$data{_nets}{$anet}{u}</A></TD><TD>$age</TD></TR>\n};
	}
	print qq{</TABLE></P>\n};
	
	#print qq{<P><TABLE BORDER=0 CELLPADDING=3 CELLSPACING=2>\n};
	#print qq{<TR><TH COLSPAN=2>Security</TH></TR>\n};
	#print qq{<TR><TH>Client</TH><TH>Age<SUP>1</SUP></TH></TR>\n};
	#foreach my $asec (@{$data{_sec}}) {
	#	my $age = getage($now-$asec->{t});
	#	print qq{<TR><TD>$asec->{h}</TD><TD>$age</TD></TR>\n};
	#}
	#print qq{</TABLE></P>\n};
	
	#warningsToBrowser(1); #DB
	
	my $age = int($security_time/60);
	print <<"(END BODY HTML)";
<!--P>
<SUP>1</SUP> Clients must wait $age minutes before returning
</P-->

<!-- Config Data
\@allowed_nets = qw(@allowed_nets);
\$default_net = '$default_net';
\$data_file = '$data_file';
\$stat_log = '$stat_log';
\$max_logsize = $max_logsize;
\$max_hosts = $max_hosts;
\$max_urls = $max_urls;
\$max_nets = $max_nets;
# Security
\$security_time = $security_time;
\@default_allow = qw(@default_allow);
# Debug Options
\$debug_no_security = $debug_no_security;
\$debug_pretend_security = $debug_pretend_security;
\$debug_out_html = $debug_out_html;
-->
<PRE>
@output
</PRE>
</BODY>
</HTML>
(END BODY HTML)
}

sub getage {
	# Input: age in seconds
	# Output: day/hr/min string
	my $a = shift;
	my $d=int($a/86400);
	my $h=int(($a%86400)/3600);
	my $m=int((($a%86400)%3600)/60);
	my $s=($d?$d.'d ':'') . ($h?$h.'h ':'') . ($m?$m.'m':'');
	return $s?$s:'0m';
}
